"""
Web application
(from web.py)
"""
import webapi as web
import webapi, wsgi, utils
import debugerror
from utils import lstrips, safeunicode
import sys

import urllib
import traceback
import itertools
import os
import types
from exceptions import SystemExit

try:
    import wsgiref.handlers
except ImportError:
    pass # don't break people with old Pythons

__all__ = [
    "application", "auto_application",
    "subdir_application", "subdomain_application", 
    "loadhook", "unloadhook",
    "autodelegate"
]

class application:
    """
    Application to delegate requests based on path.
    
        >>> urls = ("/hello", "hello")
        >>> app = application(urls, globals())
        >>> class hello:
        ...     def GET(self): return "hello"
        >>>
        >>> app.request("/hello").data
        'hello'
    """
    def __init__(self, mapping=(), fvars={}, autoreload=None):
        if autoreload is None:
            autoreload = web.config.get('debug', False)
        self.init_mapping(mapping)
        self.fvars = fvars
        self.processors = []
        
        self.add_processor(loadhook(self._load))
        self.add_processor(unloadhook(self._unload))
        
        if autoreload:
            def main_module_name():
                mod = sys.modules['__main__']
                file = getattr(mod, '__file__', None) # make sure this works even from python interpreter
                return file and os.path.splitext(os.path.basename(file))[0]

            def modname(fvars):
                """find name of the module name from fvars."""
                file, name = fvars.get('__file__'), fvars.get('__name__')
                if file is None or name is None:
                    return None

                if name == '__main__':
                    # Since the __main__ module can't be reloaded, the module has 
                    # to be imported using its file name.                    
                    name = main_module_name()
                return name
                
            mapping_name = utils.dictfind(fvars, mapping)
            module_name = modname(fvars)
            
            def reload_mapping():
                """loadhook to reload mapping and fvars."""
                mod = __import__(module_name, None, None, [''])
                mapping = getattr(mod, mapping_name, None)
                if mapping:
                    self.fvars = mod.__dict__
                    self.init_mapping(mapping)

            self.add_processor(loadhook(Reloader()))
            if mapping_name and module_name:
                self.add_processor(loadhook(reload_mapping))

            # load __main__ module usings its filename, so that it can be reloaded.
            if main_module_name() and '__main__' in sys.argv:
                try:
                    __import__(main_module_name())
                except ImportError:
                    pass
                    
    def _load(self):
        web.ctx.app_stack.append(self)
        
    def _unload(self):
        web.ctx.app_stack = web.ctx.app_stack[:-1]
        
        if web.ctx.app_stack:
            # this is a sub-application, revert ctx to earlier state.
            oldctx = web.ctx.get('_oldctx')
            if oldctx:
                web.ctx.home = oldctx.home
                web.ctx.homepath = oldctx.homepath
                web.ctx.path = oldctx.path
                web.ctx.fullpath = oldctx.fullpath
                
    def _cleanup(self):
        # Threads can be recycled by WSGI servers.
        # Clearing up all thread-local state to avoid interefereing with subsequent requests.
        utils.ThreadedDict.clear_all()

    def init_mapping(self, mapping):
        self.mapping = list(utils.group(mapping, 2))

    def add_mapping(self, pattern, classname):
        self.mapping.append((pattern, classname))

    def add_processor(self, processor):
        """
        Adds a processor to the application. 
        
            >>> urls = ("/(.*)", "echo")
            >>> app = application(urls, globals())
            >>> class echo:
            ...     def GET(self, name): return name
            ...
            >>>
            >>> def hello(handler): return "hello, " +  handler()
            ...
            >>> app.add_processor(hello)
            >>> app.request("/web.py").data
            'hello, web.py'
        """
        self.processors.append(processor)

    def request(self, localpart='/', method='GET', data=None,
                host="0.0.0.0:8080", headers=None, https=False, **kw):
        """Makes request to this application for the specified path and method.
        Response will be a storage object with data, status and headers.

            >>> urls = ("/hello", "hello")
            >>> app = application(urls, globals())
            >>> class hello:
            ...     def GET(self): 
            ...         web.header('Content-Type', 'text/plain')
            ...         return "hello"
            ...
            >>> response = app.request("/hello")
            >>> response.data
            'hello'
            >>> response.status
            '200 OK'
            >>> response.headers['Content-Type']
            'text/plain'

        To use https, use https=True.

            >>> urls = ("/redirect", "redirect")
            >>> app = application(urls, globals())
            >>> class redirect:
            ...     def GET(self): raise web.seeother("/foo")
            ...
            >>> response = app.request("/redirect")
            >>> response.headers['Location']
            'http://0.0.0.0:8080/foo'
            >>> response = app.request("/redirect", https=True)
            >>> response.headers['Location']
            'https://0.0.0.0:8080/foo'

        The headers argument specifies HTTP headers as a mapping object
        such as a dict.

            >>> urls = ('/ua', 'uaprinter')
            >>> class uaprinter:
            ...     def GET(self):
            ...         return 'your user-agent is ' + web.ctx.env['HTTP_USER_AGENT']
            ... 
            >>> app = application(urls, globals())
            >>> app.request('/ua', headers = {
            ...      'User-Agent': 'a small jumping bean/1.0 (compatible)'
            ... }).data
            'your user-agent is a small jumping bean/1.0 (compatible)'

        """
        path, maybe_query = urllib.splitquery(localpart)
        query = maybe_query or ""
        
        if 'env' in kw:
            env = kw['env']
        else:
            env = {}
        env = dict(env, HTTP_HOST=host, REQUEST_METHOD=method, PATH_INFO=path, QUERY_STRING=query, HTTPS=str(https))
        headers = headers or {}

        for k, v in headers.items():
            env['HTTP_' + k.upper().replace('-', '_')] = v

        if 'HTTP_CONTENT_LENGTH' in env:
            env['CONTENT_LENGTH'] = env.pop('HTTP_CONTENT_LENGTH')

        if 'HTTP_CONTENT_TYPE' in env:
            env['CONTENT_TYPE'] = env.pop('HTTP_CONTENT_TYPE')

        if method in ["POST", "PUT"]:
            data = data or ''
            import StringIO
            if isinstance(data, dict):
                q = urllib.urlencode(data)
            else:
                q = data
            env['wsgi.input'] = StringIO.StringIO(q)
            if not env.get('CONTENT_TYPE', '').lower().startswith('multipart/') and 'CONTENT_LENGTH' not in env:
                env['CONTENT_LENGTH'] = len(q)
        response = web.storage()
        def start_response(status, headers):
            response.status = status
            response.headers = dict(headers)
            response.header_items = headers
        response.data = "".join(self.wsgifunc()(env, start_response))
        return response

    def browser(self):
        import browser
        return browser.AppBrowser(self)

    def handle(self):
        fn, args = self._match(self.mapping, web.ctx.path)
        return self._delegate(fn, self.fvars, args)
        
    def handle_with_processors(self):
        def process(processors):
            try:
                if processors:
                    p, processors = processors[0], processors[1:]
                    return p(lambda: process(processors))
                else:
                    return self.handle()
            except web.HTTPError:
                raise
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                print >> web.debug, traceback.format_exc()
                raise self.internalerror()
        
        # processors must be applied in the resvere order. (??)
        return process(self.processors)
                        
    def wsgifunc(self, *middleware):
        """Returns a WSGI-compatible function for this application."""
        def peep(iterator):
            """Peeps into an iterator by doing an iteration
            and returns an equivalent iterator.
            """
            # wsgi requires the headers first
            # so we need to do an iteration
            # and save the result for later
            try:
                firstchunk = iterator.next()
            except StopIteration:
                firstchunk = ''

            return itertools.chain([firstchunk], iterator)    
                                
        def is_generator(x): return x and hasattr(x, 'next')
        
        def wsgi(env, start_resp):
            # clear threadlocal to avoid inteference of previous requests
            self._cleanup()

            self.load(env)
            try:
                # allow uppercase methods only
                if web.ctx.method.upper() != web.ctx.method:
                    raise web.nomethod()

                result = self.handle_with_processors()
                if is_generator(result):
                    result = peep(result)
                else:
                    result = [result]
            except web.HTTPError, e:
                result = [e.data]

            result = web.safestr(iter(result))

            status, headers = web.ctx.status, web.ctx.headers
            start_resp(status, headers)
            
            def cleanup():
                self._cleanup()
                yield '' # force this function to be a generator
                            
            return itertools.chain(result, cleanup())

        for m in middleware: 
            wsgi = m(wsgi)

        return wsgi

    def run(self, *middleware):
        """
        Starts handling requests. If called in a CGI or FastCGI context, it will follow
        that protocol. If called from the command line, it will start an HTTP
        server on the port named in the first command line argument, or, if there
        is no argument, on port 8080.
        
        `middleware` is a list of WSGI middleware which is applied to the resulting WSGI
        function.
        """
        return wsgi.runwsgi(self.wsgifunc(*middleware))
    
    def cgirun(self, *middleware):
        """
        Return a CGI handler. This is mostly useful with Google App Engine.
        There you can just do:
        
            main = app.cgirun()
        """
        wsgiapp = self.wsgifunc(*middleware)

        try:
            from google.appengine.ext.webapp.util import run_wsgi_app
            return run_wsgi_app(wsgiapp)
        except ImportError:
            # we're not running from within Google App Engine
            return wsgiref.handlers.CGIHandler().run(wsgiapp)
    
    def load(self, env):
        """Initializes ctx using env."""
        ctx = web.ctx
        ctx.clear()
        ctx.status = '200 OK'
        ctx.headers = []
        ctx.output = ''
        ctx.environ = ctx.env = env
        ctx.host = env.get('HTTP_HOST')

        if env.get('wsgi.url_scheme') in ['http', 'https']:
            ctx.protocol = env['wsgi.url_scheme']
        elif env.get('HTTPS', '').lower() in ['on', 'true', '1']:
            ctx.protocol = 'https'
        else:
            ctx.protocol = 'http'
        ctx.homedomain = ctx.protocol + '://' + env.get('HTTP_HOST', '[unknown]')
        ctx.homepath = os.environ.get('REAL_SCRIPT_NAME', env.get('SCRIPT_NAME', ''))
        ctx.home = ctx.homedomain + ctx.homepath
        #@@ home is changed when the request is handled to a sub-application.
        #@@ but the real home is required for doing absolute redirects.
        ctx.realhome = ctx.home
        ctx.ip = env.get('REMOTE_ADDR')
        ctx.method = env.get('REQUEST_METHOD')
        ctx.path = env.get('PATH_INFO')
        # http://trac.lighttpd.net/trac/ticket/406 requires:
        if env.get('SERVER_SOFTWARE', '').startswith('lighttpd/'):
            ctx.path = lstrips(env.get('REQUEST_URI').split('?')[0], ctx.homepath)
            # Apache and CherryPy webservers unquote the url but lighttpd doesn't. 
            # unquote explicitly for lighttpd to make ctx.path uniform across all servers.
            ctx.path = urllib.unquote(ctx.path)

        if env.get('QUERY_STRING'):
            ctx.query = '?' + env.get('QUERY_STRING', '')
        else:
            ctx.query = ''

        ctx.fullpath = ctx.path + ctx.query
        
        for k, v in ctx.iteritems():
            if isinstance(v, str):
                ctx[k] = safeunicode(v)

        # status must always be str
        ctx.status = '200 OK'
        
        ctx.app_stack = []

    def _delegate(self, f, fvars, args=[]):
        def handle_class(cls):
            meth = web.ctx.method
            if meth == 'HEAD' and not hasattr(cls, meth):
                meth = 'GET'
            if not hasattr(cls, meth):
                raise web.nomethod(cls)
            tocall = getattr(cls(), meth)
            return tocall(*args)
            
        def is_class(o): return isinstance(o, (types.ClassType, type))
            
        if f is None:
            raise web.notfound()
        elif isinstance(f, application):
            return f.handle_with_processors()
        elif is_class(f):
            return handle_class(f)
        elif isinstance(f, basestring):
            if f.startswith('redirect '):
                url = f.split(' ', 1)[1]
                if web.ctx.method == "GET":
                    x = web.ctx.env.get('QUERY_STRING', '')
                    if x:
                        url += '?' + x
                raise web.redirect(url)
            elif '.' in f:
                mod, cls = f.rsplit('.', 1)
                mod = __import__(mod, None, None, [''])
                cls = getattr(mod, cls)
            else:
                cls = fvars[f]
            return handle_class(cls)
        elif hasattr(f, '__call__'):
            return f()
        else:
            return web.notfound()

    def _match(self, mapping, value):
        for pat, what in mapping:
            if isinstance(what, application):
                if value.startswith(pat):
                    f = lambda: self._delegate_sub_application(pat, what)
                    return f, None
                else:
                    continue
            elif isinstance(what, basestring):
                what, result = utils.re_subm('^' + pat + '$', what, value)
            else:
                result = utils.re_compile('^' + pat + '$').match(value)
                
            if result: # it's a match
                return what, [x for x in result.groups()]
        return None, None
        
    def _delegate_sub_application(self, dir, app):
        """Deletes request to sub application `app` rooted at the directory `dir`.
        The home, homepath, path and fullpath values in web.ctx are updated to mimic request
        to the subapp and are restored after it is handled. 
        
        @@Any issues with when used with yield?
        """
        web.ctx._oldctx = web.storage(web.ctx)
        web.ctx.home += dir
        web.ctx.homepath += dir
        web.ctx.path = web.ctx.path[len(dir):]
        web.ctx.fullpath = web.ctx.fullpath[len(dir):]
        return app.handle_with_processors()
            
    def get_parent_app(self):
        if self in web.ctx.app_stack:
            index = web.ctx.app_stack.index(self)
            if index > 0:
                return web.ctx.app_stack[index-1]
        
    def notfound(self):
        """Returns HTTPError with '404 not found' message"""
        parent = self.get_parent_app()
        if parent:
            return parent.notfound()
        else:
            return web._NotFound()
            
    def internalerror(self):
        """Returns HTTPError with '500 internal error' message"""
        parent = self.get_parent_app()
        if parent:
            return parent.internalerror()
        elif web.config.get('debug'):
            import debugerror
            return debugerror.debugerror()
        else:
            return web._InternalError()

class auto_application(application):
    """Application similar to `application` but urls are constructed 
    automatiacally using metaclass.

        >>> app = auto_application()
        >>> class hello(app.page):
        ...     def GET(self): return "hello, world"
        ...
        >>> class foo(app.page):
        ...     path = '/foo/.*'
        ...     def GET(self): return "foo"
        >>> app.request("/hello").data
        'hello, world'
        >>> app.request('/foo/bar').data
        'foo'
    """
    def __init__(self):
        application.__init__(self)

        class metapage(type):
            def __init__(klass, name, bases, attrs):
                type.__init__(klass, name, bases, attrs)
                path = attrs.get('path', '/' + name)

                # path can be specified as None to ignore that class
                # typically required to create a abstract base class.
                if path is not None:
                    self.add_mapping(path, klass)

        class page:
            path = None
            __metaclass__ = metapage

        self.page = page

# The application class already has the required functionality of subdir_application
subdir_application = application
                
class subdomain_application(application):
    """
    Application to delegate requests based on the host.

        >>> urls = ("/hello", "hello")
        >>> app = application(urls, globals())
        >>> class hello:
        ...     def GET(self): return "hello"
        >>>
        >>> mapping = (r"hello\.example\.com", app)
        >>> app2 = subdomain_application(mapping)
        >>> app2.request("/hello", host="hello.example.com").data
        'hello'
        >>> response = app2.request("/hello", host="something.example.com")
        >>> response.status
        '404 Not Found'
        >>> response.data
        'not found'
    """
    def handle(self):
        host = web.ctx.host.split(':')[0] #strip port
        fn, args = self._match(self.mapping, host)
        return self._delegate(fn, self.fvars, args)
        
    def _match(self, mapping, value):
        for pat, what in mapping:
            if isinstance(what, basestring):
                what, result = utils.re_subm('^' + pat + '$', what, value)
            else:
                result = utils.re_compile('^' + pat + '$').match(value)

            if result: # it's a match
                return what, [x for x in result.groups()]
        return None, None
        
def loadhook(h):
    """
    Converts a load hook into an application processor.
    
        >>> app = auto_application()
        >>> def f(): "something done before handling request"
        ...
        >>> app.add_processor(loadhook(f))
    """
    def processor(handler):
        h()
        return handler()
        
    return processor
    
def unloadhook(h):
    """
    Converts an unload hook into an application processor.
    
        >>> app = auto_application()
        >>> def f(): "something done after handling request"
        ...
        >>> app.add_processor(unloadhook(f))    
    """
    def processor(handler):
        try:
            result = handler()
            is_generator = result and hasattr(result, 'next')
        except:
            # run the hook even when handler raises some exception
            h()
            raise

        if is_generator:
            return wrap(result)
        else:
            h()
            return result
            
    def wrap(result):
        def next():
            try:
                return result.next()
            except:
                # call the hook at the and of iterator
                h()
                raise

        result = iter(result)
        while True:
            yield next()
            
    return processor

def autodelegate(prefix=''):
    """
    Returns a method that takes one argument and calls the method named prefix+arg,
    calling `notfound()` if there isn't one. Example:

        urls = ('/prefs/(.*)', 'prefs')

        class prefs:
            GET = autodelegate('GET_')
            def GET_password(self): pass
            def GET_privacy(self): pass

    `GET_password` would get called for `/prefs/password` while `GET_privacy` for 
    `GET_privacy` gets called for `/prefs/privacy`.
    
    If a user visits `/prefs/password/change` then `GET_password(self, '/change')`
    is called.
    """
    def internal(self, arg):
        if '/' in arg:
            first, rest = arg.split('/', 1)
            func = prefix + first
            args = ['/' + rest]
        else:
            func = prefix + arg
            args = []
        
        if hasattr(self, func):
            try:
                return getattr(self, func)(*args)
            except TypeError:
                raise web.notfound()
        else:
            raise web.notfound()
    return internal

class Reloader:
    """Checks to see if any loaded modules have changed on disk and, 
    if so, reloads them.
    """

    SUFFIX = '$py.class' if sys.platform.startswith('java') else '.pyc'
    """File suffix of compiled modules."""

    def __init__(self):
        self.mtimes = {}

    def __call__(self):
        for mod in sys.modules.values():
            self.check(mod)

    def check(self, mod):
        # jython registers java packages as modules but they either
        # don't have a __file__ attribute or its value is None
        if not (mod and hasattr(mod, '__file__') and mod.__file__):
            return

        try: 
            mtime = os.stat(mod.__file__).st_mtime
        except (OSError, IOError):
            return
        if mod.__file__.endswith(self.__class__.SUFFIX) and os.path.exists(mod.__file__[:-1]):
            mtime = max(os.stat(mod.__file__[:-1]).st_mtime, mtime)
            
        if mod not in self.mtimes:
            self.mtimes[mod] = mtime
        elif self.mtimes[mod] < mtime:
            try: 
                reload(mod)
                self.mtimes[mod] = mtime
            except ImportError: 
                pass
                
if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
Model layer for the Environment's admin module.

Uses CouchDB, can create database and synchronize views during the startup.

"""

from claun.core import container
from claun.modules.couchdb import client as db
from claun.modules.couchdb import CouchError
from claun.modules.environment.model import computer_by_id

__docformat__ = 'restructuredtext en'

DB_NAME = container.configuration['modules']['environment']['db_name']
"""Default db name. Database should be created if it does not exist."""

class ModelError(Exception): pass

####

def computers_by_id_map(doc):
    """
    Maps all documents with type 'computer'.

    The generated pairs are [id: {values}].
    """
    if doc['type'] == 'computer':
        yield(doc['id'], doc)

def computers_by_masterpriority_map(doc):
    """
    Gets all computers by masterpriority.

    The generated pairs are [doc.masterpriority: {doc}].
    """
    if doc['type'] == 'computer':
        yield(doc['masterpriority'], doc)


def masterpriority_map(doc):
    """
    Gets masterpriority for all computers by id.

    This is helpful when determnining for example the max master priority.

    The generated pairs are [doc.id: doc.masterpriority].
    """
    if doc['type'] == 'computer':
        yield(doc['id'], doc['masterpriority'])

def projections_map(doc):
    """
    Lists all available projections.

    If a parameter in a configuration document has a 'group' attribute with value
    'projection', it is considered as a projection.

    Generated pairs: [projection.name, projection.name]
    """
    if doc['type'] == 'configuration':
        for p in doc['parameters']:
            if 'group' in p and p['group'] == 'projection':
                yield(p['name'], p['name'])

def projections_reduce(keys, values):
    """
    Returns only values.
    """
    return values

def raw_parameters_map(doc):
    """
    Gets raw configuration document.

    Useful for parameter deletion/update.
    """
    if doc['type'] == 'configuration':
        yield(doc['_id'], doc)

# public API

def all_nodes():
    """
    Uses computers/by_id.
    :rtype: ViewResults or []
    """
    try:
        return db.view(DB_NAME, 'computers/by_id')
    except CouchError as ce:
        container.log.error(ce)
        return []

def get_node(id):
    """
    Uses computers/by_id.

    If no such computer is found, None is returned.
    """
    result = db.view(DB_NAME, 'computers/by_id', key=id)
    try:
        if len(result) > 0:
            return result.__iter__().next()
    except Exception as e:
        container.log.debug("Node retrieval error: %s" % e)
        return None

def get_node_by_masterpriority(priority):
    """
    Uses computers/by_masterpriority.

    If no such computer is found, None is returned.
    """
    result = db.view(DB_NAME, 'computers/by_masterpriority', key=priority)
    try:
        if len(result) > 0:
            return result.__iter__().next()
    except Exception as e:
        container.log.debug("Node retrieval error: %s" % e)
        return None


def create_node(node):
    """
    Creates new node.

    If a node with given id already exists, a ModelError is raised.

    :param node: Has to contain id key.
    """
    if 'id' not in node:
        raise ModelError('Missing ID')
    if computer_by_id(node['id']) is not None:
        raise ModelError('Node already exists')
    try:
        return db.save_document(DB_NAME, node)
    except CouchError as ce:
        container.log.error(ce)
        return None

def update_node(node, docid=None, revid=None):
    """Updates node.

    You can specify `_id` and `_rev` either directly in the dictionary
    or as separate arguments. If `_id` or `_rev` are not present in either form,
    a ModelError is raised.

    :param node: Dictionary with node
    :param docid: `_id` attribute, if `node` contains `_id` field, may be None
    :param revid: `_rev` attribute, if `node` contains `_rev` field, may be None
    :return: (id, rev) tuple of the newly saved document or None
    """
    if '_id' not in node:
        if docid is None:
            raise ModelError('No ID specified!')
        node['_id'] = docid

    if '_rev' not in node:
        if revid is None:
            raise ModelError('No revision specified!')
        node['_rev'] = revid
    try:
        return db.save_document(DB_NAME, node)
    except CouchError as ce:
        container.log.error(ce)
        return None

def delete_node(node):
    """
    Deletes node, `node` has to contain _id attribute.
    """
    if '_id' not in node:
        raise ModelError('Missing _ID')
    try:
        return db.delete_document(DB_NAME, node) is None
    except CouchError as ce:
        container.log.error(ce)
        return False

def next_master_priority():
    """
    Returns next free master priority. (max + 1).

    Uses computers/max_masterpriority. May return None.
    """
    try:
        result = db.view(DB_NAME, 'computers/max_masterpriority')
        if len(result) > 0:
            return 1 + result.__iter__().next().value['max']
    except CouchError as ce:
        container.log.error(ce)
    return None

def projections():
    """
    Returns all projections available in the system.

    Uses parameters/projections.
    May return [].
    """
    try:
        result = db.view(DB_NAME, 'parameters/projections')
        if len(result) > 0:
            return result.__iter__().next().value
    except CouchError as ce:
        container.log.error(ce)
    return []

def raw_parameters():
    """
    Returns raw parameters document or None
    """
    try:
        result = db.view(DB_NAME, 'parameters/raw')
        if len(result) > 0:
            return result.__iter__().next()
    except Exception as e:
        container.log.debug("Parameters retrieval error: %s" % e)
        return None

###

def all_parameters():
    """
    Lists all parameters.

    Uses parameters/all.
    """
    try:
        return db.view(DB_NAME, 'parameters/all')
    except CouchError as ce:
        container.log.error(ce)
        return []

def get_parameter(paramname):
    """
    Uses parameters/all.

    If no such parameter is found, None is returned.
    """
    result = db.view(DB_NAME, 'parameters/all', key=paramname)
    try:
        if len(result) > 0:
            return result.__iter__().next()
    except Exception as e:
        container.log.debug("Parameter retrieval error: %s" % e)
        return None

def create_parameter(param):
    """
    Creates new parameter.

    If a parameter with given id already exists or configuration document cannot be found, a ModelError is raised.

    :param param: Has to contain id key.
    :return: (id, rev) tuple of the newly saved document
    """
    if 'id' not in param:
        raise ModelError('Missing ID')
    parameters = raw_parameters()
    try:
        if parameters is not None and 'parameters' in parameters.value:
            for p in parameters.value['parameters']:
                if p['id'] == param['id']:
                    raise ModelError('Parameter already exists')
            parameters.value['parameters'].append(param)
            return db.save_document(DB_NAME, parameters.value)
    except CouchError as ce:
        container.log.error(ce)
        return None
    raise ModelError('Cannot find parameters definition.')

def update_parameter(oldid, param):
    """Updates param.

    Because param.id might have changed, an oldid is required to
    remove the old attribute would that be the case.

    May raise ModelError if a parameter with new paramid already exists or
    a configuration document can not be found.

    :param oldid: Old parameter id
    :param param: Dict with param, has to contain id
    :return: (id, rev) tuple of the newly saved document
    """
    parameters = raw_parameters()
    try:
        if parameters is not None and 'parameters' in parameters.value:
            for idx, p in enumerate(parameters.value['parameters']):
                if p['id'] == param['id'] and oldid != param['id']:
                    raise ModelError('Parameter already exists')
                if p['id'] == oldid:
                    parameters.value['parameters'][idx] = param
                    break
            return db.save_document(DB_NAME, parameters.value)
    except CouchError as ce:
        container.log.error(ce)
        return None
    raise ModelError('Cannot find parameters definition.')

def delete_parameter(paramid):
    """
    Deletes parameter wit paramid.

    May raise ModelError if a configuration document can not be found.
    :return: True if deleted, False otherwise
    """
    parameters = raw_parameters()
    try:
        if parameters is not None and 'parameters' in parameters.value:
            for idx, p in enumerate(parameters.value['parameters']):
                if p['id'] == paramid:
                    del parameters.value['parameters'][idx]
                    return db.save_document(DB_NAME, parameters.value) is not None
    except CouchError as ce:
        container.log.error(ce)
        return False
    raise ModelError('Cannot find parameters definition.')

# sync DB
try:
    db.touch_database(DB_NAME)
    views = [
        db.define_view('computers', 'by_masterpriority', computers_by_masterpriority_map),
        db.define_view('computers', 'by_id', computers_by_id_map),
        db.define_view('computers', 'max_masterpriority', masterpriority_map, '_stats'),
        db.define_view('parameters', 'projections', projections_map, projections_reduce),
        db.define_view('parameters', 'raw', raw_parameters_map),
    ]
    db.sync_views(DB_NAME, views, False)

    if not all_parameters():
        doc = {
            "type": "configuration",
            "parameters": []
        }
        container.log.info('Creating parameters definition')
        db.save_document(DB_NAME, doc)

except CouchError as ce:
    raise container.errors.FatalError(ce)




"""
Model layer for the Applications module.

The model uses CouchDB as a storage. The application parameters can
be changed, however the changed values are not stored back
to the database, but in the personalized in-memory cache.

During startup, a database may be created and filled with views.
If views in the database are out of date, they are updated.

Every time, when querying personalized parameters, database is contacted
for current results. If the parameter was edited before by the user,
**only the edited value** is overwritten. If the parameter has
user_editable attribute set to false, its value can not be overwritten.
"""
from claun.modules.couchdb import client as db
from claun.core import container
from claun.modules.couchdb import CouchError


__docformat__ = 'restructuredtext en'
__all__ = ['all_applications', 'application_by_id', 'parameters', 'parameter_by_name', 'save_parameter', 'get_attachment']

DB_NAME = container.configuration['modules']['applications']['db_name']
"""Default db name. Database should be created if it does not exist."""

####

def visible_applications_by_id_map(doc):
    """
    All visible applications by id, gets rid of _rev, _id, visible and type attributes.

    Resulting pairs: [id: {values}]
    """
    if doc['type'] == 'application' and doc['visible']:
        notallowed = ('_rev', '_id', 'type', 'visible')
        filtered = dict((k, doc[k]) for k in doc if k not in notallowed)
        yield(filtered['id'], filtered)

def parameters_by_application_id_map(doc):
    """
    All parameters for one application.

    Resulting pair: [app.id: {app.parameters}]
    """
    if doc['type'] == 'application':
        if 'parameters' not in doc:
            yield(doc['id'], [])
        else:
            yield(doc['id'], doc['parameters'])

def parameters_by_app_and_id_map(doc):
    """
    One parameter for given app.

    Sets parameter's id from its name. Key is a tuple consisting of application id and parameter id.

    Resulting pair: [(app.id, parameter.id), parameter]
    """
    if doc['type'] == 'application' and 'parameters' in doc:
        for p in doc['parameters']:
            p['id'] = p['name']
            yield((doc['id'], p['id']), p)

# sync DB
try:
    db.touch_database(DB_NAME)
    views = [
            db.define_view('applications', 'visible_by_id', visible_applications_by_id_map),
            db.define_view('parameters', 'by_application_id', parameters_by_application_id_map),
            db.define_view('parameters', 'by_app_and_id', parameters_by_app_and_id_map),
            ]
    db.sync_views(DB_NAME, views, False) # multiple packages do this, don't remove missing
except CouchError as ce:
    raise container.errors.FatalError(ce)

# user's cache - enables per client changes
class UserCache:
    """
    Simple in-memory cache storing user edited parameters for every application.

    Derived from claun.modules.environment.model.UserCache.

    Uses dictionary as an underlying implementation.
    """
    def __init__(self):
        """Initializes the underlying dictionary."""
        self.hash = {}

    def set(self, user, appid, parameter):
        """
        Adds new parameter to the application in the user's cache. If a user does not have a cache, it is created.
        If the cache for given appid does not exist, it is created.

        Textual true|false values are transformed into True and False symbols in the parameter dictionary.

        :param user: dictionary, has to have '_id' key. Ideally, this is the raw record from couchdb.
        :param appid: Some unique application id
        :param parameter: dictionary with the parameter itself, has to have 'id' key.
        """
        if user['_id'] not in self.hash:
            self.hash[user['_id']] = {}

        if appid not in self.hash[user['_id']]:
            self.hash[user['_id']][appid] = {}

        for k, v in parameter.iteritems():
            if v == u'false':
                parameter[k] = False
            elif v == u'true':
                parameter[k] = True

        self.hash[user['_id']][appid][parameter['id']] = parameter
        return True


    def get_all(self, user, appid):
        """
        Returns all parameters saved for given user for given appid.

        If no appid cache or no user suits the parameters, None is returned.

        :param user: dictionary, has to have '_id' key. May be None, then None is returned.
        :param appid: Application id provided when setting the first parameter
        """
        if user is not None and user['_id'] in self.hash and appid in self.hash[user['_id']]:
            return self.hash[user['_id']][appid]
        return None

    def get(self, user, appid, parameter):
        """
        Returns specified parameter from specified application in user's cache.

        User and appid are passed to get_all method and then the parameter is looked up.
        If no such parameter exists in cache, None is returned.

        :param parameter: dictionary, has to have 'id' key.
        """
        conf = self.get_all(user, appid)
        if conf is None or parameter['id'] not in conf:
            return None
        else:
            return conf[parameter['id']]

# instantiate cache
cache = UserCache()

# public API
def all_applications():
    """
    All applications. Uses applications/visible_by_id view.

    :return: ViewResults or []
    """
    try:
        return db.view(DB_NAME, 'applications/visible_by_id')
    except CouchError as ce:
        container.log.error(ce)
        return []

def application_by_id(appid):
    """
    Application with id `appid`. Uses application/visible_by_id view.

    If no such application exists, None is returned.
    """
    try:
        result = db.view(DB_NAME, 'applications/visible_by_id', key=appid)
        if len(result) > 0:
            return result.__iter__().next()
    except Exception as e:
        container.log.error("Application retrieval error: %s" % e)
        return None

def parameters(user, appid):
    """
    All parameters for application `appid`. Uses parameter/by_application_id view.

    All curent data from database is returned except for parameters that can be user_editable
    and have its edited value in the user cache (They are overwritten with the cached value).

    :return: [] or None if no parameters are found for that application
    """
    try:
        result = db.view(DB_NAME, 'parameters/by_application_id', key=appid)
        if len(result) > 0:
            result = result.__iter__().next().value
            cached = cache.get_all(user, appid)
            if cached is None:
                return result

            ret = []
            for k in result:
                if k['id'] in cached and k['user_editable']: # overwrite only if edited and editable
                    k['value'] = cached[k['id']]['value'] if 'value' in cached[k['id']] else k['default']
                ret.append(k)
            return ret
        return None
    except Exception as e:
        container.log.error("Configuration retrieval error: %s" % e)
        return []

def parameter_by_name(user, appid, paramid):
    """
    One parameter with `paramid` for application `appid`.

    Current data is obtained from DB using the parameters/by_app_and_id view and
    if edited before, the current value is overwritten with the cached value.

    May return None if some Exception occurs during the process.
    """
    try:
        result = db.view(DB_NAME, 'parameters/by_app_and_id', key=(appid, paramid))
        if len(result) > 0:
            param = result.__iter__().next()
            cached = cache.get(user, appid, paramid)
            if cached is not None and param['user_editable']:
                param['value'] = cached['value'] if 'value' in cached else param['default']
            return param
    except Exception as e:
        container.log.error("Parameter retrieval error: %s" % e)
        return None

def save_parameter(user, appid, data):
    """
    Saves parameter's new value to the `user`'s `appid` cache.

    If parameter does not exist, returns False.
    :param data: Dictionary with parameter values, has to contain 'id' key.
    """
    try:
        exists = db.view(DB_NAME, 'parameters/by_app_and_id', key=(appid, data['id']))
        if len(exists) < 1:
            return False
        return cache.set(user, appid, data)
    except CouchError as ce:
        container.log.error(ce)
        return False

def get_attachment(docid, name):
    """
    Returns the attachment of `name` for the given `docid`.

    Uses couchdb.CouchWrapper.get_attachment method. See that for more details.

    :param docid: CouchDB ID of the application document that the attachment is attached to.
    :return: File like object or None
    """
    try:
        return db.get_attachment(DB_NAME, docid, name)
    except CouchError as ce:
        container.log.error(ce)
        return None



from claun.core import container

__all__ = ['ControllerClient', 'ControllerClientError']

class ControllerClientError(Exception): pass

class ControllerClient:
    """
    This class communicates with the claun.modules.controller module that can be installed on some cluster nodes.

    It is always assigned to one certain node.
    """
    def __init__(self, computer, httpclient):
        """
        Initializing of data structures.

        :param computer: Node that this instance will take care of.
        :param httpclient: claun.core.communication.client.HttpClient instance.
        """
        self.computer = computer
        self.client = httpclient

    def _available(self):
        """
        Checks instant availability/reachability of self.computer via its 'reachable' attribute.

        If computer is not reachable or does not provide controller module,
        a ControllerClientError is raised.
        """
        if not self.computer['reachable']:
            raise ControllerClientError('Unreachable node.')
        if 'controller' not in self.computer['root']['available_modules']:
            raise ControllerClientError('Missing controller module.')

    def _endpoint(self, name):
        """
        Contacts the node and tries to get an enpoints related to controller `name`.

        If the controller is not supported on the target machine, a ControllerClientError is raised.
        Otherwise link to `name` detail is returned.

        Uses `_available` method.
        """
        self._available()
        all = self.client.request(address=self.computer['root']['available_modules']['controller'])[1]['endpoints']
        if name not in all:
            raise ControllerClientError('Unsupported controller type')
        return all[name]


    def status(self, name):
        """
        Contact the `name`'s endpoint contents.

        Uses _status method.

        If the node's unreachable, None is returned.
        """
        try:
            return self._status(name)['contents']
        except container.errors.ClientException as ce:
            container.log.error("Cannot reach client %s:%i - %s" % (self.computer['data']['fqdn'], self.computer['data']['port'], ce))
            return None

    def _status(self, name):
        """
        Returns {'endpoint': url, 'contents': contents} of the `name` endpoint.

        To get `name` endpoint's address, uses self._endpoint method.
        Does not catch ClientException.
        """
        endpoint = self._endpoint(name)
        return {'endpoint': endpoint, 'contents': self.client.request(address=endpoint)[1]}

    def _process(self, name, action, configuration=''):
        """
        Sends request to `name` to do `action`.

        Returns True or parsed contents of the error response body.

        :param name: Name of the controller
        :param action: Action to perform. Supported are currently stop, start, restart. However when calling their value obtained
        from the endpoint contents is used.
        :param configuration: It should be serializable and is transported in the HTTP body under the configuration key. Only mandatory for start and restart actions.
        """
        if action not in ['start', 'stop', 'restart']:
            raise ControllerClientError('Unsupported action')

        try:
            address, status = self._status(name).values()
            if action == 'stop':
                body = container.output({'action': status['actions']['stop']}, header=container.output.default)
            else:
                body = container.output({'configuration': configuration, 'action': status['actions'][action]}, header=container.output.default)
            self.client.request(address=address, body=body, method=container.http_methods.PUT) # 202 is deemed successful
            return True
        except container.errors.ClientException as ce: # other codes than 200
            if ce.response is not None and ce.response[0]['status'] == '202':
                return True
            container.log.error("Cannot reach client %s:%i - %s" % (self.computer['data']['fqdn'], self.computer['data']['port'], ce))
            if ce.response is not None:
                return ce.response[1]
            else:
                return str(ce)



    def start(self, name, configuration):
        """
        Tries to start the `name` controller with `configuration`.

        It is not guaranteed, that the call will affect anything, as there already might be a running instance.

        Uses self._process method.
        """
        return self._process(name, 'start', configuration)


    def restart(self, name, configuration):
        """
        Tries to restart the `name` controller with `configuration`.

        In contrast to start method, when using the restart, it is guaranteed, that the target node will stop and start
        the controller program again.

        Uses self._process method.
        """
        return self._process(name, 'restart', configuration)

    def stop(self, name):
        """
        Tries to stop the `name` controller.

        Uses self._process method.
        """
        return self._process(name, 'stop')
